@@ -884,7 +884,37 @@ bool JT_PushMessage::take(const QDomElement &e)
 	if(e.tagName() != "message")
 		return false;
 
-	Stanza s = client()->stream().createStanza(addCorrectNS(e));
+	QDomElement e1 = e;
+	QDomElement forward;
+	Message::CarbonDir cd = Message::NoCarbon;
+
+	// Check for Carbon
+	QDomNodeList list = e1.childNodes();
+	for (int i = 0; i < list.size(); ++i) {
+		QDomElement el = list.at(i).toElement();
+
+		if (el.attribute("xmlns") == QLatin1String("urn:xmpp:carbons:2") && (el.tagName() == QLatin1String("received") || el.tagName() == QLatin1String("sent"))) {
+			QDomElement el1 = el.firstChildElement();
+			if (el1.tagName() == QLatin1String("forwarded") && el1.attribute(QLatin1String("xmlns")) == QLatin1String("urn:xmpp:forward:0")) {
+				QDomElement el2 = el1.firstChildElement(QLatin1String("message"));
+				if (!el2.isNull()) {
+					forward = el2;
+					cd = el.tagName() == QLatin1String("received")? Message::Received : Message::Sent;
+					break;
+				}
+			}
+		}
+		else if (el.tagName() == QLatin1String("forwarded") && el.attribute(QLatin1String("xmlns")) == QLatin1String("urn:xmpp:forward:0")) {
+			forward = el.firstChildElement(QLatin1String("message")); // currently only messages are supportted
+			// TODO <delay> element support
+			if (!forward.isNull()) {
+				break;
+			}
+		}
+	}
+
+	QString from = e1.attribute(QLatin1String("from"));
+	Stanza s = client()->stream().createStanza(addCorrectNS(forward.isNull()? e1 : forward));
 	if(s.isNull()) {
 		//printf("take: bad stanza??\n");
 		return false;
@@ -895,6 +925,10 @@ bool JT_PushMessage::take(const QDomElement &e)
 		//printf("bad message\n");
 		return false;
 	}
+	if (!forward.isNull()) {
+		m.setForwardedFrom(Jid(from));
+		m.setCarbonDirection(cd);
+	}
 
 	emit message(m);
 	return true;
@@ -2180,3 +2214,47 @@ bool JT_CaptchaSender::take(const QDomElement &x)
 
 	return true;
 }
+
+//----------------------------------------------------------------------------
+// JT_MessageCarbons
+//----------------------------------------------------------------------------
+JT_MessageCarbons::JT_MessageCarbons(Task *parent)
+	: Task(parent)
+{
+
+}
+
+void JT_MessageCarbons::enable()
+{
+	_iq = createIQ(doc(), "set", "", id());
+
+	QDomElement enable = doc()->createElement("enable");
+	enable.setAttribute("xmlns", "urn:xmpp:carbons:2");
+
+	_iq.appendChild(enable);
+}
+
+void JT_MessageCarbons::disable()
+{
+	_iq = createIQ(doc(), "set", "", id());
+
+	QDomElement disable = doc()->createElement("disable");
+	disable.setAttribute("xmlns", "urn:xmpp:carbons:2");
+
+	_iq.appendChild(disable);
+}
+
+void JT_MessageCarbons::onGo()
+{
+	send(_iq);
+	setSuccess();
+}
+
+bool JT_MessageCarbons::take(const QDomElement &e)
+{
+	if (e.tagName() != "iq" || e.attribute("type") != "result")
+		return false;
+
+	bool res = iqVerify(e, Jid(), id());
+	return res;
+}
